"""
All this file is currently in 'yta_multimedia' library
because I'm testing it, but the concepts contained in
this file don't belong to this library because they are
related to a VideoEditor concept, not to image or video
simple editing.

We need to handle, when working in building a whole video
project, videos as SubClips so we handle all attributes
and, if we subclip a SubClip instance, we .copy() the
previous attributes to the left, center and right clips
we obtain when subclipping. This would preserve previous
configurations and let us manage all the clips, so we
work on top of moviepy library in any change we process
and use moviepy only for basic and frame transformations.

TODO: These classes below will be moved in a near future
to its own project or to 'youtube_autonomous'.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.edition.settings import VOLUME_LIMIT, ZOOM_LIMIT, LAYERS_INDEXES_LIMIT, COLOR_TEMPERATURE_LIMIT, MAX_TIMELINE_LAYER_DURATION, BRIGHTNESS_LIMIT, CONTRAST_LIMIT, WHITE_BALANCE_LIMIT, SHARPNESS_LIMIT, SPEED_FACTOR_LIMIT, ROTATION_LIMIT
from yta_multimedia.video.edition.video_frame_t_helper import VideoFrameTHelper
from yta_multimedia.audio.edition.audio_frame_editor import AudioFrameEditor
from yta_multimedia.video.edition.video_frame_t_helper import SMALL_FIX_AMOUNT
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_image.edition.editor import ImageEditor
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip, concatenate_videoclips
from typing import Union

import numpy as np


END_OF_CLIP = 999999
END_OF_TIMELINE = 120
"""
The limit of the timeline length. It is not possible
to generate a project which length is larger than 
this value.

TODO: This value could be very low in alpha and beta
versions for testing.
"""

# TODO: Move this decorator below to another place (?)
def unset_video_processed(func):
    """
    Decorator function that sets the '_video_processed'
    attribute to None to indicate that it has been
    modified and it must be processed again.
    """
    def wrapper(self, value):
        value = func(self, value)
        self._video_processed = None
        
        return value
    
    return wrapper


from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from yta_general_utils.math.progression import Progression


class SubClipAttributeModifier:
    """
    Class to encapsulate the possible ways to modify a 
    video attribute.

    This will be passed as an instance to a SubClip to 
    calculate the modification values array for each 
    frame and set those modifying values in the video
    instance to be lately applied to the video when
    processing it.

    This is a wrapper to simplify the way we interact
    with different object types valid to generate
    values.

    The only accepted values by now are:
    - Single value
    - SubClipSetting
    - Graphic
    """
    modifier: Union[any, 'SubClipSetting', 'Graphic'] = None

    @property
    def is_single_value(self) -> bool:
        """
        Return True if the modifier is just a single value.
        """
        return PythonValidator.is_number(self.modifier)

    def __init__(self, modifier: Union[any, 'SubClipSetting', 'Graphic']):
        # TODO: Change SubClipSetting name as its purpose is
        # now different and not to appropiate for this
        if (
            # TODO: Maybe we accept some non-numeric modifier
            # single values (?)
            not PythonValidator.is_number(modifier) and
            not PythonValidator.is_instance(modifier, 'SubClipSetting') and
            not PythonValidator.is_instance(modifier, 'Graphic')
        ):
            raise Exception('The provided "modifier" parameter is not a valid modifier.')
        
        self.modifier = modifier

    def get_values(self, n: float):
        """
        Obtain an array of 'n' values that will modify the
        attribute this instance is designed for. The 'n'
        value must be the number of frames.
        """
        # I don't like float 'fps' but it is possible, and I
        # should force any clip to be 30 or 60fps always
        if PythonValidator.is_number(self.modifier): return [self.modifier] * n
        if PythonValidator.is_instance(self.modifier, 'SubClipSetting'): return self.modifier.get_values(n)
        if PythonValidator.is_instance(self.modifier, 'Graphic'): return [self.modifier.get_xy_from_normalized_d(d)[1] for d in Progression(0, 1, n).values]

    def validate_values(self, n: float, limit: list[float, float]):
        """
        Validate that any of those 'values' is between the
        limit range. The 'n' parameter must be the number of
        frames in the video, and 'limit' a tuple of the lower
        and upper limit.

        This method must be called when a SubClipAttributeModifier
        instance is set in a SubClip because we know the number
        of frames and the limit for that specific attribute (it
        is being added to that attribute modifier) in that 
        moment.
        """
        if any(value < limit[0] or value > limit[1] for value in self.get_values(n)):
            raise Exception(f'One of the generated "values" is out of the limit [{limit[0]}, {limit[1]}]')

    def copy(self):
        return SubClipAttributeModifier(self.modifier.copy())


# TODO: Think again about this because now we have the
# SubClipAttributeModifier to be passed as the modifier
# and it accepts this SubClipSetting, that must be renamed
# for its new purpose
class SubClipSetting:
    """
    Class to represent a video setting to be able to handle
    dynamic setting values and not only simple values. This
    means we can make a video go from 0 contrast to 10 
    contrast increasing it smoothly (for example: 0, 2, 4,
    6, 8 and 10) and not only abruptly (from 0 in one frame
    to 10 in the next frame).
    """
    initial_value: float = None
    final_value: float = None
    rate_function: RateFunctionArgument = None
    _values: list[float] = None
    """
    The list of values calculated in a previous method call
    to avoid recalculating them. The amount of elements is
    the amount of steps used in the previous calculation.
    """

    def __init__(
        self,
        initial_value: float,
        final_value: float,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        # TODO: Maybe validate something? I don't know the
        # limits because each setting is different, but by
        # now I'm verifying the 'initial_value' and the
        # 'final_value' when using them on a SubClip
        self.initial_value = initial_value
        self.final_value = final_value
        self.rate_function = rate_function

    def get_values(self, steps: int):
        """
        Obtain an array with the values between the 'initial_value'
        and the 'final_value' according to the 'rate_function'.

        The 'steps' parameter must be the amount of frames in which
        we are planning to apply this setting so we are able to read
        the value for each frame according to its index.
        """
        # Same limits cannot be handled by the Progression class as
        # it is just an array of the same value repeated 'steps' 
        # times
        if self.initial_value == self.final_value:
            return [self.initial_value for _ in range(steps)]
        
        if self._values is None or len(self._values) != steps:
            # We recalculate only if needed
            self._values = Progression(self.initial_value, self.final_value, steps, self.rate_function).values
            
        return self._values
    
    def copy(self):
        return SubClipSetting(
            self.initial_value,
            self.final_value,
            self.rate_function
        )


class SubClip:
    """
    Class to represent a subclip of a clip in which we
    can apply different modifications such as color
    temperature, zoom, movement, etc.

    This class represent the same as one of the subclips
    in any of your video editor apps.
    """
    # TODO: This, as the original video clip (moviepy),
    # maybe must be private to avoid using it directly
    video: Clip = None
    _video_processed: Clip = None
    """
    The video once it's been processed with all its
    attributes and effects.
    """
    # TODO: Add all needed attributes
    # Volume
    _volume: list[float] = None
    """
    A single value or a list of values for the volume 
    modification in which the whole video audio volume
    is modified by that single factor or those factor
    values.
    """

    # Attributes that can be modified
    _color_temperature: list[float] = None
    """
    A list of values for the color temperature modification
    in which each position is the modifier for each video
    frame.
    """
    _brightness: list[float] = None
    """
    A list of values for the brightness modification in
    which each position is the modifier for each video
    frame.
    """
    _contrast: list[float] = None
    """
    A list of values for the contrast modification in which
    each position is the modifier for each video frame.
    """
    _sharpness: list[float] = None
    """
    A list of values for the sharpness modification in which
    each position is the modifier for each video frame.
    """
    _white_balance: list[float] = None
    """
    A list of values for the white balance modification in
    which each position is the modifier for each video frame.
    """

    # Special modifiers
    # TODO: This is special because it affects to the video
    # duration. It has to be applied at the end
    _speed_factor: Union[float, list[float]] = None
    """
    A single value or a list of values for the speed 
    modification in which the whole video duration is
    modified by that single speed factor or those speed
    factor values.
    """

    # Zoom, movement and rotation. These attributes are
    # single values that are changed the base clip size,
    # position and rotation.
    _zoom: float = None
    """
    A single value to zoom the clip.
    """
    _x_movement: float = None
    """
    A single value to move the clip in X axis.
    """
    _y_movement: float = None
    """
    A single value to move the clip in Y axis.
    """
    _rotation: float = None
    """
    A single value to rotate the clip in clockhise.
    """

    # These 3 values below are the most especial ones
    # because are the ones that actually build the
    # video by applying other effects or attributes
    # on themselves
    _resized: list[tuple[float, float]] = None
    """
    A list of tuples with the width and height resize
    factors that must be applied for each frame to 
    the original video.
    """
    _with_position: list[tuple[int, int]] = None
    """
    A list of positions in which the center of the
    video must be placed. These positions must be
    transformed into upper left corner positions just
    when processing and rendering.
    """
    _rotated: list[int] = None
    """
    A list of rotation angles for each frame of the
    video.

    The rotation effect makes a frame with a new size
    so it has to be considered when positioning to
    calculate the actual size of each frame.
    """
    
    # Custom effects
    _effects: list['SEffect'] = None

    @staticmethod
    def init(video: Clip, start_time: Union[float, None] = 0, end_time: Union[float, None] = END_OF_CLIP):
        """
        This is the only method we need to use when instantiating
        a SubClip so we can obtain the left and right clip result
        of the subclipping process and also the new SubClip
        instance.

        This method returns a tuple with 3 elements, which are the
        left part of the subclipped video, the center part and the
        right part as SubClip instances. If no left or right part,
        they are return as None. So, the possibilities are (from
        left to right):

        - SubClip, SubClip, SubClip
        - SubClip, SubClip, None
        - None, SubClip, SubClip
        - None, SubClip, SubClip
        """
        video = VideoParser.to_moviepy(video)

        # We use a flag to indicate the end of the clip so
        # we replace it by its actual duration
        end_time = video.duration if end_time == END_OF_CLIP else end_time

        left_clip, center_clip, right_clip = subclip_video(video, start_time, end_time)

        return (
            SubClip(left_clip) if left_clip is not None else None,
            SubClip(center_clip),
            SubClip(right_clip) if right_clip is not None else None
        )

    def __init__(self, video: Clip):
        """
        I recommend you to use the static '.init()'method 
        instead of this because you are able to keep the 
        remaining part of the clip if you subclip them.

        The SubClip instantiating method has to be called only
        in the static 'init' method of this class so we are able
        to handle the rest of the clips (if existing) according
        to the subclipping process we do. If you use this method
        and then subclip
        """
        self.video = VideoParser.to_moviepy(video)
        
        self._speed_factor = 1
        self._volume = 1
        self._zoom = 1
        self._x_movement = 0
        self._y_movement = 0
        self._rotation = 0
        # TODO: This was being calculated by VideoHandler
        # to ensure the number of frames was ok
        self._resized = [(1, 1) for _ in range(self.number_of_frames)]
        self._rotated = [0 for _ in range(self.number_of_frames)]
        # TODO: Use constants
        self._with_position = [(MOVIEPY_SCENE_DEFAULT_SIZE[0] / 2, MOVIEPY_SCENE_DEFAULT_SIZE[1] / 2) for _ in range(self.number_of_frames)]
        self._effects = []

    @property
    def video_processed(self):
        """
        The video once it's been processed with all the
        attributes that are set in this instance.
        """
        if not self._video_processed:
            self._video_processed = self._process()

        return self._video_processed

    @property
    def duration(self):
        """
        Shortcut to the actual duration of the video once
        it's been processed.
        """
        # TODO: Maybe this has to be the pre-processed
        #return self.video_processed.duration
        return self.video.duration
    
    @property
    def original_size(self):
        """
        The size of the original video clip.
        """
        return self.video.size

    @property
    def original_width(self):
        """
        The width of the original video clip.
        """
        return self.original_size[0]
    
    @property
    def original_height(self):
        """
        The height of the original video clip.
        """
        return self.original_size[1]

    @property
    def size(self):
        """
        Shortcut to the actual size of the video once
        it's been processed.
        """
        return self.video_processed.size
    
    @property
    def width(self):
        """
        Shortcut to the actual width of the processed
        video.
        """
        return self.size[0]
    
    @property
    def height(self):
        """
        Shortcut to the actual height of the processed
        video.
        """
        return self.size[1]
    
    @property
    def fps(self):
        """
        Fps of the original video (the pre-processed one).
        """
        return self.video.fps
    
    @property
    def number_of_frames(self):
        """
        Number of frames of the original video (the 
        pre-processed one).

        Due to a bug with the moviepy reading way, this
        number cannot be the real number. Check the 
        MPVideo class that looks for the exact number.
        """
        # TODO: Maybe use MPVideo to ensure the video
        # has the real number of frames, and also 
        # subclip the video to the real duration based
        # on this. I have a method that does it by 
        # trying to get the last frames and catching
        # the warnings
        return int(self.duration * self.fps + SMALL_FIX_AMOUNT)
    
    @property
    def number_of_audio_frames(self):
        """
        Number of frames of the original audio (the
        pre-processed one).
        """
        return int(self.video.audio.duration * self.video.audio.fps + SMALL_FIX_AMOUNT)
    
    @property
    def volume(self):
        return self._volume
    
    @volume.setter
    def volume(self, value: SubClipAttributeModifier):
        """
        Set the video audio volume values by providing
        a SubClipAtrributeModifier that will set those
        values for this SubClip instance.

        TODO: Say the limit values and explain more.
        """
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'volume', VOLUME_LIMIT, self.number_of_frames)

        # Values from 0 to 100 are for humans, but this
        # is a factor so we normalize it
        self._volume = value.modifier / 100 if value.is_single_value else [value / 100 for value in value.get_values(self.number_of_frames)]

    @property
    def color_temperature(self):
        return self._color_temperature
    
    @color_temperature.setter
    def color_temperature(self, value: SubClipAttributeModifier):
        """
        Set the color temperature values by providing a 
        SubClipAttributeModifier that will set the values
        for the current SubClip.
        """
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'color_temperature', COLOR_TEMPERATURE_LIMIT, self.number_of_frames)

        self._color_temperature = value.get_values(self.number_of_frames)

    @property
    def brightness(self):
        return self._brightness
    
    @brightness.setter
    def brightness(self, value: SubClipAttributeModifier):
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'brightness', BRIGHTNESS_LIMIT, self.number_of_frames)

        self._brightness = value.get_values(self.number_of_frames)

    @property
    def contrast(self):
        return self._contrast
    
    @contrast.setter
    def contrast(self, value: SubClipAttributeModifier):
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'contrast', CONTRAST_LIMIT, self.number_of_frames)

        self._contrast = value.get_values(self.number_of_frames)

    @property
    def sharpness(self):
        return self._sharpness

    @sharpness.setter
    def sharpness(self, value: SubClipAttributeModifier):
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'sharpness', SHARPNESS_LIMIT, self.number_of_frames)

        self._sharpness = value.get_values(self.number_of_frames)

    @property
    def white_balance(self):
        return self._white_balance
    
    @white_balance.setter
    def white_balance(self, value: SubClipAttributeModifier):
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'white_balance', WHITE_BALANCE_LIMIT, self.number_of_frames)

        self._white_balance = value.get_values(self.number_of_frames)

    @property
    def speed_factor(self):
        return self._speed_factor
    
    @speed_factor.setter
    def speed_factor(self, value: SubClipAttributeModifier):
        _validate_is_video_attribute_modifier_instance(value)
        _validate_attribute_modifier(value, 'speed_factor', SPEED_FACTOR_LIMIT, self.number_of_frames)

        self._speed_factor = value.modifier if value.is_single_value else value.get_values(self.number_of_frames) 

    @property
    def zoom(self):
        return self._zoom
    
    @zoom.setter
    @unset_video_processed
    def zoom(self, value: int):
        """
        The zoom must be an integer number between [1, 500]. A
        zoom value of 100 means no zoom.
        
        - Zoom=100 means no zoom
        - Zoom=50 means zoom out until the clip size is the half.
        - Zoom=200 means zoom in until the clip size is the doubel
        """
        _validate_zoom(value)

        self._zoom = value / 100

    @property
    def x_movement(self):
        return self._x_movement
    
    @x_movement.setter
    def x_movement(self, value: int):
        """
        The movement in X axis must be a value between
        [-1920*4, 1920*4]. A positive number means moving
        the clip to the right side.
        """
        _validate_x_movement(value)
        
        self._x_movement = int(value)

    @property
    def y_movement(self):
        return self._y_movement
    
    @y_movement.setter
    def y_movement(self, value: int):
        """
        The movement in Y axis must be a value between
        [-1920*4, 1920*4]. A positive number means moving
        the clip to the bottom.
        """
        _validate_y_movement(value)
        
        self._y_movement = int(value)

    @property
    def rotation(self):
        return self._rotation
    
    @rotation.setter
    def rotation(self, value: int):
        """
        Rotation must be an integer value between -360
        and 360. A positive number means rotating the
        clip clockwise.
        """
        _validate_rotation(value)
        
        self._rotation = int(value)

    @property
    def effects(self):
        return self._effects

    # Easy setters below, that are another way of setting
    # attributes values but just passing arguments. This
    # is interesting if you just need to apply a simple
    # and single value or an easy range
    def set_volume(
        self,
        start: float,
        end: Union[float, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new volume for the video that will be modified frame by
        frame.

        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('volume', start, end, rate_function)

    def set_color_temperature(
        self,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new color temperature that will be modified frame by frame.

        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('color_temperature', start, end, rate_function)

    def set_brightness(
        self,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new brightness that will be modified frame by frame.
        
        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('brightness', start, end, rate_function)

    def set_contrast(
        self,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new contrast that will be modified frame by frame.
        
        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('contrast', start, end, rate_function)

    def set_sharpness(
        self,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new sharpness that will be modified frame by frame.
        
        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('sharpness', start, end, rate_function)

    def set_white_balance(
        self,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new white balance that will be modified frame by frame.
        
        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('white_balance', start, end, rate_function)

    def set_speed_factor(
        self,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Set a new speed factor that will be modifier frame by frame.

        If only 'start' is provided, the change will be the same in all
        frames, but if a different 'end' value is provided, the also given
        'rate_function' will be used to calculate the values in between
        those 'start' and 'end' limits to be applied in the corresponding
        frames.
        """
        self._set_attribute('speed_factor', start, end, rate_function)

    def _set_attribute(
        self,
        attribute: str,
        start: int,
        end: Union[int, None] = None,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        setattr(
            self,
            attribute,
            SubClipAttributeModifier(
                SubClipSetting(
                    start,
                    start if end is None else end,
                    rate_function
                )
            )
        )

    # Other easy setters
    def set_zoom(self, value: int):
        """
        Set a new zoom value.
        """
        self.zoom = value

    def set_x_movement(self, value: int):
        """
        Set a new x movement value.
        """
        self.x_movement = value

    def set_y_movement(self, value: int):
        """
        Set a new y movement value.
        """
        self.y_movement = value

    def set_rotation(self, value: int):
        """
        Set a new rotation value.
        """
        self.rotation = value

    def add_effect(self, effect: 'SEffect'):
        """
        Add the provided 'effect' instance to be applied on the clip.
        """
        if not PythonValidator.is_an_instance(effect) or not PythonValidator.is_subclass(effect, 'SEffect'):
            raise Exception('The provided "effect" parameter is not an instance of a SEffect subclass.')
        
        # TODO: Check that effect is valid (times are valid, there is
        # not another effect that makes it incompatible, etc.)
        # We force the correct 'number_of_frames' attribute
        effect.number_of_frames = self.number_of_frames

        self._effects.append(effect)

    def subclip(self, start: float, end: float):
        """
        This method will split the current SubClip instance
        into 3 different items, that will be new instances
        of SubClip class according to the 'start' and 'end'
        times provided.

        This method uses a copy of the current instance to
        not modify it but returning completely new (by
        using deepcopy). All settings and effects will be
        preserved as they were in the original instance for
        all of the new copies.

        This method will return 3 values: left part of the
        SubClip, center part and right part. Left and right
        part can be None.
        """
        # TODO: Validate
        if not NumberValidator.is_positive_number(start):
            raise Exception('The "start" parameter provided is not a valid and positive number (0 included).')
        
        if not NumberValidator.is_positive_number(end):
            raise Exception('The "end" parameter provided is not a valid and positive number (0 included).')
        
        if start >= end:
            raise Exception('The "start" parameter provided is greater or equal than the "end" parameter provided.')
        
        if end > self.duration:
            raise Exception(f'The "end" provided ({end}s) is longer than the current video duration ({self.duration}s).')

        # TODO: I need to calculate the frame index in which I'm
        # splitting the subclip to also subclip the arrays that
        # are inside the instance
        left = self.copy() if start not in (None, 0) else None
        center = self.copy()
        right = self.copy() if end is not None and end < self.duration else None

        def replace_attribute_values(instance: SubClip, start_index: Union[float, None], end_index: Union[float, None]):
            """
            Replaces the attribute values of the given 'instance' 
            considering the left, center and right videos that will
            be returned as result so each video can keep only their
            values.
            """
            def split_array(array: list, start_index: Union[float, None], end_index: Union[float, None]):
                # TODO: Validate 'start_index' and 'end_index' according
                # to 'array' size
                if start_index is not None and end_index is not None: return array[start_index:end_index]
                if start_index is not None: return array[start_index:]
                if end_index is not None: return array[:end_index]
                return array

            # TODO: Append here any new array of values per frame
            instance._color_temperature = split_array(instance._color_temperature, start_index, end_index) if instance._color_temperature is not None else None
            instance._brightness = split_array(instance._brightness, start_index, end_index) if instance._brightness is not None else None
            instance._contrast = split_array(instance._contrast, start_index, end_index) if instance._contrast is not None else None
            instance._sharpness = split_array(instance._sharpness, start_index, end_index) if instance._sharpness is not None else None
            instance._white_balance = split_array(instance._white_balance, start_index, end_index) if instance._white_balance is not None else None
            # TODO: '_speed_factor' is special and can be both a single
            # value or a list (or a None if no changes to apply)
            if PythonValidator.is_list(instance._speed_factor):
                instance._speed_factor = split_array(instance._speed_factor, start_index, end_index)
            
            return instance

        last_index = 0
        # Left
        if left is not None:
            left.video = left.video.with_subclip(0, start)
            last_index = left.number_of_frames
            # Modify all the attributes per frame values
            left = replace_attribute_values(left, 0, last_index)

        # Center
        center.video = center.video.with_subclip(start, end)
        # Modify all the attributes per frame values
        center = replace_attribute_values(center, last_index, last_index + center.number_of_frames)
        last_index = last_index + center.number_of_frames

        # Right
        if right is not None:
            right.video = right.video.with_subclip(start_time = end)
            # Modify all the attributes per frame values
            right = replace_attribute_values(right, last_index, None)

        return left, center, right
    
    def write_videofile(self, output_filename: str):
        """
        Writes the video once it is processed to the provided
        'output_filename'.
        """
        # TODO: Validate 'output_filename'

        self.video_processed.write_videofile(output_filename)

    def copy(self):
        # TODO: Complete this method to manually copy the instance
        # because 'deepcopy' is not working properly
        copy = SubClip(self.video.copy())

        # The only thing we need to preserve is the values that
        # modify each attribute. The modifier instance is only
        # passed to generate these values, so that generator is
        # only necessary once to generate those values
        copy._color_temperature = self._color_temperature.copy() if self._color_temperature is not None else None
        copy._brightness = self._brightness.copy() if self._brightness is not None else None
        copy._contrast = self._contrast.copy() if self._contrast is not None else None
        copy._sharpness = self._sharpness.copy() if self._sharpness is not None else None
        copy._white_balance = self._white_balance.copy() if self._white_balance is not None else None
        copy._speed_factor = self._speed_factor.copy() if self._speed_factor is not None else None

        copy._zoom = self._zoom if self._zoom is not None else None
        copy._x_movement = self._x_movement if self._x_movement is not None else None
        copy._y_movement = self._y_movement if self._y_movement is not None else None
        copy._rotation = self._rotation if self._rotation is not None else None

        # TODO: Maybe I need to apply the basic values here (?)
        copy._resized = self._resized.copy() if self._resized is not None else None
        copy._rotated = self._rotated.copy() if self._rotated is not None else None
        copy._with_position = self._with_position.copy() if self._with_position is not None else None

        return copy
    
    def _process(self):
        """
        Process the video clip with the attributes set and 
        obtain a copy of the original video clip with those
        attributes and effects applied on it. This method
        uses a black (but transparent) background with the
        same video size to make sure everything works 
        properly.

        This method doesn't change the original clip, it
        applies the changes on a copy of the original one
        and returns that copy modified.
        """
        video = self.video.copy()

        # Process video frames one by one
        def modify_video_frame_by_frame(get_frame, t):
            """
            Modificate anything related to frame image: pixel
            colors, distortion, etc.
            """
            frame = get_frame(t)
            frame_index = VideoFrameTHelper.get_frame_index_from_frame_t(t, video.fps)

            frame = ImageEditor.modify_color_temperature(frame, self._color_temperature[frame_index]) if self._color_temperature is not None else frame
            frame = ImageEditor.modify_brightness(frame, self._brightness[frame_index]) if self._brightness is not None else frame
            frame = ImageEditor.modify_contrast(frame, self._contrast[frame_index]) if self._contrast is not None else frame
            frame = ImageEditor.modify_sharpness(frame, self._sharpness[frame_index]) if self._sharpness is not None else frame
            frame = ImageEditor.modify_white_balance(frame, self._white_balance[frame_index]) if self._white_balance is not None else frame

            return frame
        
        # Apply frame by frame video modifications
        video = video.transform(lambda get_frame, t: modify_video_frame_by_frame(get_frame, t))

        # Process audio frames one by one
        def modify_audio_frame_by_frame(get_frame, t):
            # The 't' time moment is here a numpy array
            # of a lot of consecutive time moments (maybe
            # 1960). That ishow it works internally
            frame = get_frame(t)
            frame_index = VideoFrameTHelper.get_video_frame_index_from_video_audio_frame_t(t[500], video.fps, video.audio.fps)

            # Volume value 1 means no change
            if self._volume is not None and self._volume != 1:
                frame = AudioFrameEditor.modify_volume(
                    frame,
                    self._volume if PythonValidator.is_number(self._volume) else self._volume[frame_index]
                )
                # TODO: The result with this below is the same
                #frame *= self._volume if PythonValidator.is_number(self._volume) else self._volume[frame_index]

            return frame
        
        if video.audio is not None:
            audio = video.audio.copy()
            audio = audio.transform(lambda get_frame, t: modify_audio_frame_by_frame(get_frame, t))
            video = video.with_audio(audio)

        # Edit speed with speed factors (carefully)
        if PythonValidator.is_number(self.speed_factor) and self.speed_factor != 1:
            from yta_multimedia.video.edition.effect.fit_duration_effect import FitDurationEffect
            video = FitDurationEffect().apply(video, self.duration / self.speed_factor)
        elif PythonValidator.is_list(self.speed_factor):
            # TODO: 'resizes', 'positions' and 'rotations' must be
            # refactored also according to the new fps
            video = self._apply_speed_factor(video)

        # The '_apply_speed_factor' updates these arrays
        resizes = self._resized.copy()
        positions = self._with_position.copy()
        rotations = self._rotated.copy()

        # Modifications below affect to other important
        # attributes.
        # TODO: I don't know if I will accept these
        # modifications below in any case or if I will
        # block them if there are some effects or things
        # that can make conflicts or malfunctioning. If
        # I have a very specific effect, changing these
        # important attributes (resizes, positions and
        # rotations) could be a big headache.
        if self.zoom is not None and self.zoom != 1:
            resizes = [(resize[0] * self.zoom, resize[1] * self.zoom) for resize in resizes]
        if self.x_movement is not None and self.x_movement != 0:
            positions = [(position[0] + self.x_movement, position[1]) for position in positions]
        if self.y_movement is not None and self.y_movement != 0:
            positions = [(position[0], position[1] + self.y_movement) for position in positions]
        if self.rotation is not None and self.rotation != 0:
            rotations = [rotation + self.rotation for rotation in rotations]

        # TODO: Should we apply effects here after the
        # general basic attributes modifications and after
        # speed factor is applied, or before? That's a 
        # a good question that only testing can answer
        if len(self.effects) > 0:
            self._apply_effects()

        """
        The rotation process makes a redimension of the
        frame image setting the pixels out of the rotation
        as alpha to actually build this effect. Thats why
        we need to know the new frame size to be able to
        position it correctly in the position we want
        """
        # Apply the video rotation, frame by frame
        video = video.rotated(lambda t: rotations[VideoFrameTHelper.get_frame_index_from_frame_t(t, video.fps)], expand = True)

        # As the rotation changes the frame size, we need
        # to recalculate the resize factors
        # TODO: Move this method to a helper or something
        def get_rotated_image_size(size: tuple[int, int], angle: int):
            """
            Get the size of an image of the given 'size' when it
            is rotated the also given 'angle'.

            This method is based on the moviepy Rotate effect to
            pre-calculate the frame rotation new size so we are
            able to apply that resize factor to the other 
            attributes.

            This method returns the new size and also the width
            size change factor and the height size change factor.
            """
            from PIL import Image

            new_size = Image.new('RGB', size, (0, 0, 0)).rotate(
                angle,
                expand = True,
                resample = Image.Resampling.BILINEAR
            ).size

            width_factor = new_size[0] / size[0]
            height_factor = new_size[1] / size[1]

            return new_size, width_factor, height_factor
        
        for index, resize in enumerate(resizes):
            current_rotation = rotations[index]
            if current_rotation != 0:
                # Recalculate the resize according to the resize
                # factor that took place when rotating the image
                _, width_factor, height_factor = get_rotated_image_size(
                    (
                        int(self.original_width * resize[0]),
                        int(self.original_height * resize[1])
                    ),
                    current_rotation
                )
                
                resizes[index] = (
                    resize[0] * width_factor,
                    resize[1] * height_factor
                )

        """
        The video 'resized' method doesn't accept double
        factors so I have to manually calculate the new
        size according to the video size and pass that
        exact new size to be actually resized as I need.
        """
        # Resize the video dynamically frame by frame
        def resized(t):
            """
            Resizes the video by applying the resize factors to
            the original size and returns the new size to be
            applied.
            """
            current_frame_index = VideoFrameTHelper.get_frame_index_from_frame_t(t, video.fps)

            return (
                resizes[current_frame_index][0] * self.original_width,
                resizes[current_frame_index][1] * self.original_height
            )

        video = video.resized(lambda t: resized(t))

        """
        The position is very special because it depends on
        the size of each frame, that can change dynamically
        because of resizes, rotations, etc. The 'positions'
        array is pointing the center coordinate of the 
        position, so we need to recalculate the upper left
        corner according to that position.
        """

        def with_position(t):
            # This is very slow but I know the exact frame
            # size so I'm sure the positioning will be ok
            #frame = video.get_frame(t)
            #frame_size = frame.shape[1], frame.shape[0]
            current_frame_index = VideoFrameTHelper.get_frame_index_from_frame_t(t, video.fps)
            
            # Adjust position to its upper left corner
            upper_left_corner_position = [
                positions[current_frame_index][0] - self.width * resizes[current_frame_index][0] / 2,
                positions[current_frame_index][1] - self.height * resizes[current_frame_index][1] / 2
            ]

            return upper_left_corner_position[0], upper_left_corner_position[1]
        
        video = video.with_position(lambda t: with_position(t))

        # TODO: This below is repeated in VideoEditor class as
        # '._overlay_video()'
        return CompositeVideoClip([
            ClipGenerator.get_default_background_video(duration = video.duration, fps = video.fps),
            video
        ])#.with_audio(VideoAudioCombinator(audio_mode).process_audio(background_video, video))

    def _apply_speed_factor(self, video):
        """
        Apply the speed factors to the video. This method 
        will use the 'time_transform' method and also will
        set a new duration with the 'with_duration' method.

        This method returns the new video modified.
        """
        if PythonValidator.is_number(self.speed_factor) or self.speed_factor is None:
            raise Exception(f'The "speed_factor" parameter is not valid for this method. It must be an array of {self.number_of_frames} elements.')
        
        def _get_frame_ts_applying_speed_factors(self: SubClip):
            """
            Returns a tuple with the video and audio arrays
            affected by the speed factors.
            """
            if self.number_of_frames != len(self.speed_factor):
                raise Exception(f'The number of video frames {self.number_of_frames} and speed factors array {len(self.speed_factor)} must be the same.')

            # We only work with speed factors for video frames
            # as we know what is the audio associated to each
            # frame and can modify it according to it
            final_video_frame_ts = []
            final_audio_frame_ts = []
            # TODO: I need to have any array that has to be
            # recalculated according to the repeated or the
            # skipped indexes
            positions = []
            resizes = []
            rotations = []

            rest = 0
            current_frame_index = 0
            while current_frame_index < self.number_of_frames:
                current_speed_factor = self.speed_factor[current_frame_index]
                current_video_frame_t = VideoFrameTHelper.get_frame_t_from_frame_index(current_frame_index, self.video.fps)

                # We know current video frame 't', so we can obtain
                # the associated audio frame tts
                current_audio_frame_ts = VideoFrameTHelper.get_video_audio_tts_from_video_frame_t(current_video_frame_t, self.video.fps, self.video.audio.fps)

                if current_speed_factor < 1:
                    # If needed, repeat frames to slow down
                    times_to_append = 1
                    current_rest = (1 / current_speed_factor) - 1
                    rest -= current_rest

                    if rest <= -1:
                        times_to_append += int(abs(rest))
                        rest += int(abs(rest))
                    
                    final_video_frame_ts.extend([current_video_frame_t] * times_to_append)
                    """
                    We have two different ways of handling the audio
                    duplication.
                    
                    - One is duplicating block by block, so:
                    from [1, 2] to [1, 2, 1, 2, 1, 2]

                    - The other one duplicates value by value, so:
                    from [1, 2] to [1, 1, 1, 2, 2, 2]
                    """
                    # Block by block
                    #final_audio_frame_ts.extend([current_audio_frame_ts] * times_to_append)
                    # Value by value
                    final_audio_frame_ts.extend([
                        caft
                        for caft in current_audio_frame_ts
                        for _ in range(times_to_append)
                    ])

                    # Any other array that also has to be recalculated
                    resizes.extend([self._resized[current_frame_index]] * times_to_append)
                    positions.extend([self._with_position[current_frame_index]] * times_to_append)
                    rotations.extend([self._rotated[current_frame_index]] * times_to_append)

                else:
                    # Extend the video and the audio just once
                    final_video_frame_ts.append(current_video_frame_t)
                    final_audio_frame_ts.extend(current_audio_frame_ts)

                    # Any other array that also has to be recalculated
                    resizes.append(self._resized[current_frame_index])
                    positions.append(self._with_position[current_frame_index])
                    rotations.append(self._rotated[current_frame_index])

                    if current_speed_factor > 1:
                        # If needed, we skip frames to speed it up
                        current_rest = current_speed_factor - 1
                        rest += current_rest

                        if rest >= 1:
                            current_frame_index += int(rest)
                            rest -= int(rest)

                current_frame_index += 1

            return final_video_frame_ts, final_audio_frame_ts, resizes, positions, rotations

        final_video_frame_ts, final_audio_frame_ts, resizes, positions, rotations = _get_frame_ts_applying_speed_factors(self)

        # TODO: Remove this
        print(f'Total of {len(final_video_frame_ts)} ({len(final_video_frame_ts) / self.video.fps}) video frame ts and {len(final_audio_frame_ts)} ({len(final_audio_frame_ts) / self.video.audio.fps}) audio frame ts.')

        def transform_t_with_both_frames(t, video_fps: float, audio_fps: float):
            """
            Transform the time moment 't' we are processing
            to render in the new file according to the time
            moments of the original video/audio. We have
            pre-calculated them so we know what frame of the
            original video/audio has to be placed for each
            rendering time moment 't' we handle here.
            """
            if not PythonValidator.is_numpy_array(t):
                # Video frame. The 't' is just a time moment
                return final_video_frame_ts[
                    VideoFrameTHelper.get_frame_index_from_frame_t(t, video_fps)
                ]
            else:
                # Audio frame. The 't' is an array of time
                # moments. The amount of 't' is unespecific.
                # I think it's just a chunk. Mine was 1960
                t_indexes = [
                    VideoFrameTHelper.get_frame_index_from_frame_t(t_, audio_fps)
                    for t_ in t
                ]

                # I have to return an array that replaces
                # the original tts
                return np.array([final_audio_frame_ts[t_index] for t_index in t_indexes])

        # TODO: What if video has no audio (?)
        video = video.time_transform(
            lambda t: transform_t_with_both_frames(t, self.video.fps, self.video.audio.fps), apply_to = ['mask', 'audio']
        )
        #video = video.with_duration(len(final_video_frame_ts) * 1 / video.fps + 0.000001)
        video = video.with_duration(len(final_video_frame_ts) * 1 / video.fps)

        # TODO: I don't want this to be done here, please
        # return it and do in another place in this class
        self._resized = resizes
        # TODO: Positions must be the center of the 
        # position we want, and then calculate the upper
        # left corner when actually positioning it
        self._with_position = positions
        self._rotated = rotations

        return video

    def _apply_effects(self):
        """
        Apply the effects.
        """
        if len(self._effects) > 0:
            # TODO: Apply effects
            for effect in self._effects:
                if effect.do_affect_frames:
                    frames = effect.values[0]
                    # TODO: Handle 'frames' array, by now I'm just
                    # replacing the values
                    # TODO: Add or replace (?)
                    pass

                if effect.do_affect_with_position:
                    with_position = effect.values[1]
                    # TODO: Handle 'with_position' array, by now I'm just
                    # replacing the values
                    self._with_position = with_position
                    # TODO: Add or replace (?)

                if effect.do_affect_resized:
                    resized = effect.values[2]
                    # TODO: Handle 'resized' array, by now I'm just
                    # replacing the values
                    self._resized = resized
                    # TODO: Multiply or replace (?)

                if effect.do_affect_rotated:
                    rotated = effect.values[3]
                    # TODO: Handle 'rotated' array, by now I'm just
                    # replacing the values
                    self._rotated = rotated
                    # TODO: Add or replace (?)
    

class SubClipOnTimelineLayer:
    """
    Class to represent one of our SubClips but in
    the general project timeline and in a specific
    layer of it, with the start and end moment in
    that timeline, and also the layer in which it
    is placed.

    TODO: This is a concept in test phase. SubClip
    is a more definitive concept.
    """
    subclip: SubClip = None
    start_time: float = None
    """
    The start time on the general project timeline. Do 
    not confuse this term with the start time of a
    moviepy clip.
    """

    @property
    def video_processed(self):
        return self.subclip.video_processed

    @property
    def duration(self):
        """
        Shortcut to the actual duration of the video once
        it's been processed.
        """
        return self.subclip.duration
    
    @property
    def size(self):
        """
        Shortcut to the actual size of the video once it's
        been processed.
        """
        return self.subclip.size
    
    @property
    def width(self):
        """
        Shortcut to the actual weight of the video once
        it's been processed.
        """
        return self.size[0]
    
    @property
    def height(self):
        """
        Shortcut to the actual height of the video once
        it's been processed.
        """
        return self.size[1]
    
    @property
    def end_time(self):
        """
        The end moment on the timeline, based on this
        instance 'start_time' and the real video 'duration'.
        """
        return self.start_time + self.duration

    def __init__(self, subclip: SubClip, start_time: float):
        if not PythonValidator.is_instance(subclip, SubClip):
            raise Exception('The provided "subclip" parameter is not a valid SubClip instance.')
        
        if not NumberValidator.is_number_between(start_time, 0, END_OF_TIMELINE):
            raise Exception(f'The provided "start_time" parameter is not a valid number between in the range (0, {END_OF_TIMELINE})')
        
        self.subclip = subclip
        self.start_time = start_time

    # def _process(self):
    #     return self.subclip._process()

class TimelineLayerType(Enum):
    """
    The type of a timeline layer, which will determine
    which kind of SubClips are accepted by the layer
    with this type.
    """
    VIDEO = 'video'
    """
    The type of layer that only accept video SubClips.
    """
    AUDIO = 'audio'
    """
    The type of layer that only accept audio SubClips.
    """
    # TODO: Probably add 'GreenscreenLayer',
    # 'AlphascreenLayer', 'TextLayer', 'SubtitleLayer',
    # and all needed in a future when this concept 
    # evolves properly

class TimelineLayer:
    index: int = None
    type: TimelineLayerType = None
    _subclips: list[SubClipOnTimelineLayer] = None

    @property
    def subclips(self):
        """
        The list of subclips in this timeline layer, ordered
        according to its 'start_time' from first to be displayed
        to last ones.
        """
        return sorted(self._subclips, key = lambda subclip: subclip.start_time)
    
    @property
    def duration(self):
        """
        The duration of this timeline layer, which is the
        'end_time' of the last displayed subclip, or 0 if no
        clips.
        """
        return self.subclips[-1].end_time if self.subclips else 0
    
    @property
    def all_clips(self):
        """
        The list of all clips needed to fulfill the timeline
        layer completely. This involves the actual subclips
        but also the needed black background clips to put in
        the gaps between the subclips.
        """
        all_subclips = []
        current_time_moment = 0
        # TODO: This 'timeline_duration' must be calculated
        # according to all timeline layers, so the longest
        # one is the one we need to assign here
        timeline_duration = self.subclips[-1].end_time

        for subclip in self.subclips:
            # We fulfill the gap if existing
            if subclip.start_time > current_time_moment:
                all_subclips.append(ClipGenerator.get_default_background_video(size = subclip.size, duration = subclip.start_time - current_time_moment, fps = subclip.fps))
            
            # We add the existing subclip
            all_subclips.append(subclip.video_processed)
            
            current_time_moment = subclip.end_time
        
        # Check if gap at the end due to other longer layers
        if current_time_moment < timeline_duration:
            # TODO: Maybe we need to do something with the fps
            all_subclips.append(ClipGenerator.get_default_background_video(size = subclip.size, duration = timeline_duration - current_time_moment, fps = self.subclips[0].fps))
        
        return all_subclips

    def __init__(self, index: int = 0, type: TimelineLayerType = TimelineLayerType.VIDEO):
        _validate_layer_index(index)
        type = TimelineLayerType.to_enum(type) if type is not None else TimelineLayerType.VIDEO

        self.index = index
        self.type = type
        self._subclips = []

    def add_subclip(self, subclip: SubClip, start_time: float):
        """
        Append the provided 'subclip' at the end of the list.

        TODO: Being in the end of the list doesn't mean being
        the last one displayed. By now I'm storing them just
        one after another and ordering them when trying to get
        them as a property. This will change in a near future
        to be more eficient.
        """
        if not PythonValidator.is_instance(subclip, SubClip):
            raise Exception('The provided "subclip" parameter is not a valid instance of the SubClip class.')
        
        if not NumberValidator.is_number_between(start_time, 0, MAX_TIMELINE_LAYER_DURATION):
            raise Exception('The provided "start_time" is not a valid value')

        # 
        # TODO: Check that the 'start_time' or the 'duration'
        # doesn't collide with another existing subclip. If
        # yes, choose what strategy to follow
        if any(subclip.start_time <= start_time <= subclip.end_time for subclip in self.subclips):
            raise Exception(f'There is one existing subclip at the {str(start_time)} time position.')

        self._subclips.append(SubClipOnTimelineLayer(subclip, start_time))

    def remove_subclip(self, index: int):
        """
        Delete the subclip in the provided 'index' position of
        the list (if existing), or raises an Exception if it
        doesn't exist or the list is empty.
        """
        # TODO: Maybe remove by passing the instance (?)
        if PythonValidator.is_empty_list(self._subclips):
            # TODO: Maybe I should not raise an Exception here...
            raise Exception('No subclips to remove.')
        
        if not NumberValidator.is_number_between(0, len(self._subclips)):
            raise Exception(f'The provided "index" is not a valid index (must be between [0, {str(len(self._subclips))}).')

        # TODO: Be very careful, I have a 'subclips' property which
        # returns the subclips ordered, but the raw '_subclips'
        # property is not ordered, so here the 'index' is potentially
        # wrong. Think how to handle this in a near future, please.
        del self._subclips[index]

    def build(self):
        """
        Concatenate all the timeline layer subclips (fulfilling the
        gaps with black transparent background clips) and return the
        concatenated clip.
        """
        # TODO: What if I have one non-transparent and transparent
        # clips in this timeline layer? They will be treated in a
        # similar way so it is not the expected behaviour...
        return concatenate_videoclips(self.all_clips)
    
class VideoProject:
    """
    Class representing a whole but single video project
    in which we have a timeline with different layers
    and clips on them.
    """
    timeline_layers: list[TimelineLayer] = None
    screen_size: tuple[int, int] = None
    """
    Dimensions of the final video that has to be
    exported.
    """

    def __init__(self, screen_size: tuple[int, int] = (MOVIEPY_SCENE_DEFAULT_SIZE[0], MOVIEPY_SCENE_DEFAULT_SIZE[1])):
        self.timeline_layers = [
            TimelineLayer(0, TimelineLayerType.VIDEO),
            # TODO: I Simplify everything by now and I only
            # handle one video layer
            #TimelineLayer(0, TimelineLayerType.AUDIO)
        ]
        # TODO: Validate 'screen_size'
        self.screen_size = screen_size

    def get_layers(self, type: TimelineLayerType) -> list[TimelineLayer]:
        """
        Get the timeline layers of the provided 'type' sorted by
        index in ascending order.
        """
        type = TimelineLayerType.to_enum(type)

        return sorted(
            [layer for layer in self.timeline_layers if layer.type == type], 
            key = lambda layer: layer.index
        )
    
    def get_last_layer_index(self, type: TimelineLayerType) -> Union[int, None]:
        """
        Get the last index used for the layers of the given
        'type', that will be None if no layers of that type.
        """
        type = TimelineLayerType.to_enum(type)

        layers = self.get_layers(type)

        return layers[-1:].index if not PythonValidator.is_empty_list(layers) else None

    def add_layer(self, type: TimelineLayerType = TimelineLayerType.VIDEO) -> int:
        """
        Add a new layer of the provided 'type' and returns
        the index in which it has been placed.
        """
        type = TimelineLayerType.to_enum(type) if type is not None else TimelineLayerType.VIDEO

        layers = self.get_layers(type)
        index = len(layers) + 1 if layers else 0

        self.timeline_layers.append(TimelineLayer(index, type))

        return index
    
    def remove_layer(self, layer: TimelineLayer):
        if not PythonValidator.is_instance(layer, TimelineLayer):
            raise Exception('The provided "layer" is not an instance of the TimelineLayer class.')
        
        if layer not in self.timeline_layers:
            raise Exception('The provided "layer" does not exist in this project.')

        self.timeline_layers.remove(layer)

    def build(self):
        # TODO: Omg, cyclic import issue here again...
        from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode
        # TODO: Remove this code below when this is done 
        # individually by each layer. By now I'm forcing
        # all layers clips to have the same duration as
        # the longest layer clip has, but must be done
        # using a general and common 'max_duration'
        # property that is not calculated here.
        layers_clips = [
            layer.build()
            for layer in self.timeline_layers
        ]
        max_duration = max(layer_clip.duration for layer_clip in layers_clips)
        layers_clips = [
            set_video_duration(layer_clip, max_duration, extend_mode = ExtendVideoMode.BLACK_TRANSPARENT_BACKGROUND)
            for layer_clip in layers_clips
        ]

        return CompositeVideoClip(layers_clips)





def subclip_video(video: Clip, start_time: float, end_time: float) -> tuple[Union[Clip, None], Clip, Union[Clip, None]]:
    """
    Subclip the provided 'video' into 3 different subclips,
    according to the provided 'start_time' and 'end_time',
    and return them as a tuple of those 3 clips. First and
    third clip could be None.

    The first clip will be None when 'start_time' is 0, and 
    the third one when the 'end_time' is equal to the given
    'video' duration.
    """
    video = VideoParser.to_moviepy(video)

    left = None if (start_time == 0 or start_time == None) else video.with_subclip(0, start_time)
    center = video.with_subclip(start_time, end_time)
    right = None if (end_time is None or end_time >= video.duration) else video.with_subclip(start_time = end_time)

    return left, center, right



# TODO: This must be maybe moved to another file
# because it is mixed with the 'subclip_video'
# method...
def _validate_layer_index(layer_index: int):
    if not NumberValidator.is_number_between(layer_index, LAYERS_INDEXES_LIMIT[0], LAYERS_INDEXES_LIMIT[1]):
        raise Exception(f'The provided "layer_index" is not a valid layer, it must be an int value between [{LAYERS_INDEXES_LIMIT[0]}, {LAYERS_INDEXES_LIMIT[1]}].')
    
def _validate_is_video_attribute_modifier_instance(element: SubClipAttributeModifier):
    if not PythonValidator.is_instance(element, SubClipAttributeModifier):
        raise Exception('The provided "element" parameter is not a SubClipAttributeModifier instance.')
    
def _validate_zoom(zoom: int):
    if not NumberValidator.is_number_between(zoom, ZOOM_LIMIT[0], ZOOM_LIMIT[1]):
        raise Exception(f'The "zoom" parameter provided is not a number between [{ZOOM_LIMIT[0]}, {ZOOM_LIMIT[1]}].')
    
def _validate_x_movement(x_movement: int):
    if not NumberValidator.is_number_between(x_movement, -MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4, MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4):
        raise Exception(f'The "x_movement" parameter provided is not a number between [{-MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4}, {MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4}].')
    
def _validate_y_movement(y_movement: int):
    if not NumberValidator.is_number_between(y_movement, -MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4, MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4):
        raise Exception(f'The "y_movement" parameter provided is not a number between [{-MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4}, {MOVIEPY_SCENE_DEFAULT_SIZE[0] * 4}].')

def _validate_rotation(rotation: int):
    if not NumberValidator.is_number_between(rotation, ROTATION_LIMIT[0], ROTATION_LIMIT[1]):
        raise Exception(f'The "rotation" parameter provided is not a number between [{ROTATION_LIMIT[0]}, {ROTATION_LIMIT[1]}].')
    
def _validate_setting(setting: SubClipSetting, name: str, range: tuple[float, float]):
    if not PythonValidator.is_instance(setting, SubClipSetting):
        raise Exception(f'The provided "{name}" is not a SubClipSetting instance.')
    
    if not NumberValidator.is_number_between(setting.initial_value, range[0], range[1]):
        raise Exception(f'The "{name}" parameter provided "initial_value" is not a number between [{range[0]}, {range[1]}].')
    
    if not NumberValidator.is_number_between(setting.final_value, range[0], range[1]):
        raise Exception(f'The "{name}" parameter provided "final_value" is not a number between [{range[0]}, {range[1]}].')

def _validate_attribute_modifier(attribute_modifier: SubClipAttributeModifier, name: str, limit_range: tuple[float, float], number_of_frames: int):
    """
    Validate the provided 'attribute_modifier' according to
    the given 'limit_range' in which all the values must fit.
    Also, if it is a Graphic instance, the 'number_of_frames'
    will be used to generate the values and check them.
    """
    if not PythonValidator.is_instance(attribute_modifier, SubClipAttributeModifier):
        raise Exception(f'The parameter "{name}" provided is not a SubClipAttributeModifier instance.')
    
    if PythonValidator.is_list(attribute_modifier.modifier):
        # TODO: Validate all values
        if any(not NumberValidator.is_number_between(value, limit_range[0], limit_range[1]) for value in attribute_modifier.modifier):
            raise Exception(f'The parameter "{name}" provided has at least one value out of the limits [{limit_range[0]}, {limit_range[1]}]')
    elif PythonValidator.is_instance(attribute_modifier.modifier, SubClipSetting):
        if not NumberValidator.is_number_between(attribute_modifier.modifier.initial_value, limit_range[0], limit_range[1]):
            raise Exception(f'The parameter "{name}" provided "initial_value" is not a number between [{limit_range[0]}, {limit_range[1]}].')
        
        if not NumberValidator.is_number_between(attribute_modifier.modifier.final_value, limit_range[0], limit_range[1]):
            raise Exception(f'The parameter "{name}" provided "final_value" is not a number between [{limit_range[0]}, {limit_range[1]}].')
    elif PythonValidator.is_instance(attribute_modifier.modifier, 'Graphic'):
        # TODO: This is very agressive, according to the way
        # we join the pairs of nodes we could get outliers
        # that are obviously out of the limit range. Some
        # easing functions have values below 0 and over 1.
        if any(not NumberValidator.is_number_between(value, limit_range[0], limit_range[1]) for value in attribute_modifier.get_values(number_of_frames)):
            raise Exception(f'The parameter "{name}" provided has at least one value out of the limits [{limit_range[0]}, {limit_range[1]}]')


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyArgument, MoviepyWith, WithPositionArgument
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from moviepy.Clip import Clip
from typing import Union


class MoveLinearPositionEffect(Effect):
    """
    Move from A to B doing a straight line effect while moving.
    """
    # TODO: Is this working (?)
    result_must_replace = True

    def apply(cls, video: Union[Clip, str], initial_position: Union[Coordinate, Position], final_position: Union[Coordinate, Position]):
        video = VideoParser.to_moviepy(video)

        # TODO: Validate and parse 'initial_position' and 
        # 'final_position'. These positions must be coordinates
        # on a 1920x1080 scene.
        arg = MoviepyArgument(initial_position, final_position, TFunctionSetPosition.linear, RateFunctionArgument.default())

        return MoviepyWith.apply(video, with_position = arg)
    
    def apply_over_video(cls, video: Union[Clip, str], background_video: Union[Clip, str], initial_position: Union[Coordinate, Position], final_position: Union[Coordinate, Position]):
        arg = MoviepyArgument(initial_position, final_position, TFunctionSetPosition.linear, RateFunctionArgument.default())

        return MoviepyWith.apply_over_video(video, background_video, with_position = arg)
    
    # TODO: Should be static (?)
    def get_arrays(
        self,
        number_of_frames: int,
        fps: float,
        initial_position: Union[Coordinate, Position, tuple[int, int]],
        final_position: Union[Coordinate, Position, tuple[int, int]]
    ):
        """
        Get the 'with_position', 'resized', and 'rotated' arrays
        to apply the effect on a SubClip instance. Each array can
        be None if not affected by this effect.
        """
        from yta_general_utils.math.map.map import Map
        from yta_general_utils.programming.parameter_validator import PythonValidator

        if PythonValidator.is_instance(initial_position, Coordinate):
            initial_position = initial_position.as_moviepy_tuple
        elif PythonValidator.is_instance(initial_position, Position):
            initial_position = initial_position.get_moviepy_center_tuple()
        elif PythonValidator.is_tuple(initial_position) and len(initial_position) == 2:
            # Default scene of 1920x1080
            initial_position = initial_position
        else:
            raise Exception('The provided "initial_position" parameter is not valid.')

        if PythonValidator.is_instance(final_position, Coordinate):
            final_position = final_position.as_moviepy_tuple
        elif PythonValidator.is_instance(final_position, Position):
            final_position = final_position.get_moviepy_center_tuple() 
        elif PythonValidator.is_tuple(final_position) and len(final_position) == 2:
            # Default scene of 1920x1080
            final_position = final_position
        else:
            raise Exception('The provided "final_position" parameter is not valid.')

        return MoviepyWith.get_arrays(
            number_of_frames,
            fps,
            with_position = WithPositionArgument(Map().add_coordinate(initial_position).add_coordinate(final_position))
        )


from yta_general_utils.programming.enum import YTAEnum as Enum
from abc import ABC, abstractmethod
from typing import Tuple, Union

import numpy as np


FRAMES_TO_CHECK = 2
"""
The amount of frames we use to check what arrays
are being affected by the effect.
"""

# TODO: Move this class to its own file
class SEffectType(Enum):
    """
    Type of SEffects that we are able to handle, useful to
    recognize if we are trying to apply two (or more) 
    effects of the same type and use only one of them.

    TODO: I don't know how to indicate the type correctly
    so I can decide I'm only able to use one of that type.
    """
    GENERAL = 'general'
    """
    I don't know which type is this effect and I'm not
    blocking it.
    """
    MOVEMENT = 'movement'
    """
    Effect that only affects to 'with_position' array.
    """

class SEffect(ABC):
    """
    Class to apply and effect and return its 'with_position',
    'rotated' and 'resized' parameters.

    This class must be implemented by any specific and 
    custom effect that applies to a SubClip instance.
    """
    type: SEffectType = None
    """
    The type of the effect.
    """
    _number_of_frames: int = None
    _values: tuple[list, list, list, list] = None
    """
    The array containing the 4 arrays of values that have
    been calculated in a previous iteration. These values
    are stored to avoid recalculating them.
    """

    def __init__(
        self,
        number_of_frames: int,
        type: SEffectType = SEffectType.GENERAL,
        *args,
        **kwargs
    ):
        type = SEffectType.to_enum(type)

        self._number_of_frames = number_of_frames
        self._values = None
        self.type = type
        self.args = args
        self.kwargs = kwargs

    @property
    def number_of_frames(self):
        return self._number_of_frames
    
    @number_of_frames.setter
    def number_of_frames(self, value: int):
        """
        Set the number of frames, useful if we need to
        recalculate the values according to this new
        value. This method will force the effect values
        to be recalculated.
        """
        # TODO: Validate (?)
        self._number_of_frames = value
        # We reset the values so they have to be calculated
        # again due to 'number_of_frames' change that makes
        # them different from the previous calculation
        self._values = None

    @property
    def do_affect_frames(self) -> bool:
        """
        Return True if the effect affects to the 'frames'
        array or not.
        """
        return self.calculate(FRAMES_TO_CHECK, *self.args, **self.kwargs)[0] is not None

    @property
    def do_affect_with_position(self) -> bool:
        """
        Return True if the effect affects to the 'with_position'
        array or not.
        """
        return self.calculate(FRAMES_TO_CHECK, *self.args, **self.kwargs)[1] is not None

    @property
    def do_affect_resized(self) -> bool:
        """
        Return True if the effect affects to the 'resized'
        array or not.
        """
        return self.calculate(FRAMES_TO_CHECK, *self.args, **self.kwargs)[2] is not None

    @property
    def do_affect_rotated(self) -> bool:
        """
        Return True if the effect affects to the 'rotated'
        array or not.
        """
        return self.calculate(FRAMES_TO_CHECK, *self.args, **self.kwargs)[3] is not None

    @property
    def values(self) -> Tuple[Union[list[np.ndarray], None], Union[list[int, int], None], Union[list[float, float], None], Union[int, None]]:
        """
        Calculate the values for the 4 arrays 'frames',
        'with_position', 'resized', and 'rotated' and return
        each of them if affected, or None if not.
        """
        if self._values is None:
            self._values = self.calculate(self.number_of_frames, *self.args, **self.kwargs)

        return self._values
    
    @staticmethod
    @abstractmethod
    def calculate(
        number_of_frames: int,
        *args,
        **kwargs
    ) -> Tuple[Union[list[np.ndarray], None], Union[list[int, int], None], Union[list[float, float], None], Union[int, None]]:
        """
        Calculate the values for the 4 arrays:
         
        - frames
        - with_position
        - resized
        - rotated
        
        with the provided 'number_of_frames' and
        additional arguments and return each of
        them if affected, or None if not.
        """
        return (
            None,
            None,
            None,
            None
        )


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.programming.parameter_validator import NumberValidator
from moviepy import Clip, clips_array

import math


class MultipliedEffect(Effect):
    """
    Makes the provided 'video' appear repeated
    in the scene in 'rows' rows and columns, so
    the video appears the pow of the 'times' you
    provide.

    The larger the 'rows' number is, the longer
    it takes to render.
    """

    def apply(self, video: Clip, times: int) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        if not NumberValidator.is_positive_number(times):
            raise Exception('The provided "times" value is not a positive number.')

        # TODO: This limitation is hardcoded
        if times > 6:
            raise Exception('Sorry, to large to render... Limited until we find a way to speed up the process.')

        times *= times

        audio = video.audio
        size = (video.w, video.h)

        # TODO: Try video.resized(1 / times)
        videos_array = [
            [video] * times
            for _ in range(times)
        ]

        return clips_array(videos_array).resized(size).with_audio(audio)

    # TODO: I don't need this
    def apply_over_video(self, video, background_video):
        return super().apply_over_video(video, background_video)


"""
Nice help: https://www.bannerbear.com/blog/how-to-use-ffmpeg-in-python-with-examples/
Official doc: https://www.ffmpeg.org/ffmpeg-resampler.html
More help: https://kkroening.github.io/ffmpeg-python/
Nice guide: https://img.ly/blog/ultimate-guide-to-ffmpeg/
Available flags: https://gist.github.com/tayvano/6e2d456a9897f55025e25035478a3a50

Interesting usage: https://stackoverflow.com/a/20325676
Maybe avoid writting on disk?: https://github.com/kkroening/ffmpeg-python/issues/500#issuecomment-792281072
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.audio.parser import AudioParser
from yta_multimedia.video.dimensions import get_video_size
from yta_multimedia.utils.resize import get_cropping_points_to_keep_aspect_ratio
from yta_multimedia.video.position import validate_size
from yta_image.parser import ImageParser
from yta_general_utils.file.handler import FileHandler
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.file.writer import FileWriter
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.date import seconds_to_hh_mm_ss
from yta_general_utils.programming.parameter_validator import PythonValidator, NumberValidator
from typing import Union
from subprocess import run


class FfmpegAudioCodec(Enum):
    """
    TODO: Fill this

    Should be used in the **-c:a {codec}** flag.
    """
    AAC = 'aac'
    """
    Default encoder.
    """
    AC3 = 'ac3'
    """
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-ac3-and-ac3_005ffixed
    """
    AC3_FIXED = 'ac3_fixed'
    """
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-ac3-and-ac3_005ffixed
    """
    FLAC = 'flac'
    """
    FLAC (Free Lossless Audio Codec) Encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-flac-2
    """
    OPUS = 'opus'
    """
    This is a native FFmpeg encoder for the Opus format. Currently, it's
    in development and only implements the CELT part of the codec. Its
    quality is usually worse and at best is equal to the libopus encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-opus
    """
    LIBFDK_AAC = 'libfdk_aac'
    """
    libfdk-aac AAC (Advanced Audio Coding) encoder wrapper. The libfdk-aac
    library is based on the Fraunhofer FDK AAC code from the Android project.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libfdk_005faac
    """
    LIBLC3 = 'liblc3'
    """
    liblc3 LC3 (Low Complexity Communication Codec) encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-liblc3
    """
    LIBMP3LAME = 'libmp3lame'
    """
    LAME (Lame Ain't an MP3 Encoder) MP3 encoder wrapper.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libmp3lame-1
    """
    LIBOPENCORE_AMRNB = 'libopencore_amrnb'
    """
    OpenCORE Adaptive Multi-Rate Narrowband encoder.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libopencore_002damrnb-1ss
    """
    LIBOPUS = 'libopus'
    """
    libopus Opus Interactive Audio Codec encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libopus-1
    """
    LIBSHINE = 'libshine'
    """
    Shine Fixed-Point MP3 encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libshine-1
    """
    LIBTWOLAME = 'libtwolame'
    """
    TwoLAME MP2 encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libtwolame
    """
    LIBVO_AMRWBENC = 'libvo-amrwbenc'
    """
    VisualOn Adaptive Multi-Rate Wideband encoder.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libvo_002damrwbenc
    """
    LIBVORBIS = 'libvorbis'
    """
    libvorbis encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libvorbis
    """
    MJPEG = 'mjpeg'
    """
    Motion JPEG encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-mjpeg
    """
    WAVPACK = 'wavpack'
    """
    WavPack lossless audio encoder.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-wavpack
    """
    COPY = 'copy'
    """
    Indicates that the codec must be copied from 
    the input.
    """

class FfmpegVideoCodec(Enum):
    """
    These are the video codecs available as Enums. The amount of codecs
    available depends on the ffmpeg built version.
    
    Should be used in the **-c:v {codec}** flag.
    """
    A64_MULTI = 'a64_multi'
    """
    A64 / Commodore 64 multicolor charset encoder. a64_multi5 is extended with 5th color (colram).

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-a64_005fmulti_002c-a64_005fmulti5
    """
    A64_MULTI5 = 'a64_multi5'
    """
    A64 / Commodore 64 multicolor charset encoder. a64_multi5 is extended with 5th color (colram).

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-a64_005fmulti_002c-a64_005fmulti5
    """
    CINEPAK = 'Cinepak'
    """
    Cinepak aka CVID encoder. Compatible with Windows 3.1 and vintage MacOS.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-Cinepak
    """
    GIF = 'GIF'
    """
    GIF image/animation encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-Cinepak
    """
    HAP = 'Hap'
    """
    Vidvox Hap video encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-Hap
    """
    JPEG2000 = 'jpeg2000'
    """
    The native jpeg 2000 encoder is lossy by default

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-jpeg2000
    """
    LIBRAV1E = 'librav1e'
    """
    rav1e AV1 encoder wrapper.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-librav1e
    """
    LIBAOM_AV1 = 'libaom-av1'
    """
    libaom AV1 encoder wrapper.
    
    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libaom_002dav1
    """
    # TODO: Continue with this (https://www.ffmpeg.org/ffmpeg-codecs.html#toc-libsvtav1)
    QTRLE = 'qtrle'
    """
    TODO: Find information about this video codec.

    More info: ???
    """
    PRORES = 'prores'
    """
    Apple ProRes encoder.

    More info: https://www.ffmpeg.org/ffmpeg-codecs.html#toc-ProRes
    """
    COPY = 'copy'
    """
    Indicates that the codec must be copied from 
    the input.
    """

class FfmpegVideoFormat(Enum):
    """
    Enum list to simplify the way we choose a video format for
    the ffmpeg command. This should be used with the FfmpegFlag
    '-f' flag that forces that video format.

    Should be used in the **-f {format}** flag.
    """
    CONCAT = 'concat'
    """
    The format will be the concatenation.
    """
    AVI = 'avi'
    """
    Avi format.

    # TODO: Explain more
    """
    PNG = 'png'
    """
    # TODO: Look for mor information about this vcodec
    # TODO: I don't know if this one is actually an FfmpegVideoFormat
    # or if I need to create another Enum class. This option us used
    # in the '-vcodec' option, and the other ones are used in the
    # 'c:v' option.
    """
    # TODO: Keep going

class FfmpegFilter(Enum):
    """
    Enum list to simplify the way we use a filter for the
    ffmpeg command.

    Should be used in the **-filter {filter}** flag.
    """
    THUMBNAIL = 'thumbnail'
    """
    Chooses the most representative frame of the video to be used
    as a thumbnail.
    """

class FfmpegPixelFormat(Enum):
    """
    Enum list to simplify the way we use a pixel format for
    the ffmpeg command.

    Should be used in the **-pix_fmt {format}** flag.
    """
    YUV420p = 'yuv420p'
    """
    This is de default value. TODO: Look for more information about it
    """
    RGB24 = 'rgb24'
    """
    TODO: Look for more information about this pixel format.
    """
    ARGB = 'argb'
    """
    TODO: Look for more information about this pixel format
    """
    YUVA444P10LE = 'yuva444p10le'
    """
    TODO: Look for more information about this pixel format
    """

class FfmpegFlag:
    """
    Class to simplify the way we push flags into the ffmpeg command.
    """
    overwrite: str = '-y'
    """
    Overwrite the output file if existing.

    Notation: **-y**
    """

    @staticmethod
    def force_format(format: FfmpegVideoFormat):
        """
        Force the output format to be the provided 'format'.

        Notation: **-f {format}**
        """
        format = FfmpegVideoFormat.to_enum(format).value

        return f'-f {format}'
    
    @staticmethod
    def safe_routes(value: int):
        """
        To enable or disable unsafe paths.

        Notation: **-safe {value}**
        """
        # TODO: Check that 'value' is a number between -1 and 1

        return f'-safe {str(value)}'
    
    @staticmethod
    def input(input: str):
        """
        To set the input (or inputs) we want.

        Notation: **-i {input}**
        """
        # TODO: I don't know how to check this or format it from 'input' param

        return f'-i {input}'
    
    @staticmethod
    def audio_codec(codec: Union[FfmpegAudioCodec, str]):
        """
        Sets the general audio codec.

        Notation: **-c:a {codec}**
        """
        # We cannot control the big amount of audio codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegAudioCodec.to_enum(codec).value
        except:
            pass

        return f'-c:a {codec}'
    
    @staticmethod
    def video_codec(codec: Union[FfmpegVideoCodec, str]):
        """
        Sets the general video codec.

        Notation: **-c:v {codec}**
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegVideoCodec.to_enum(codec).value
        except:
            pass

        return f'-c:v {codec}'

    @staticmethod
    def v_codec(codec: Union[FfmpegVideoCodec, str]):
        """
        Sets the video codec.

        TODO: I don't know exactly the difference between '-c:v {codec}'
        and the '-vcodec' generated in this method. I keep this method
        until I actually find the difference. I don't even know if the
        video codecs I can provide as values are the same as in the other
        method.

        Notation: **-vcodec {codec}**
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter
        try:
            codec = FfmpegVideoCodec.to_enum(codec).value
        except:
            pass

        return f'-vcodec {codec}'

    @staticmethod
    def codec(codec: Union[FfmpegVideoCodec, FfmpegAudioCodec, str]):
        """
        Sets the general codec with '-c {codec}'.

        -c copy indica que se deben copiar los flujos de audio y video sin recodificación, lo que hace que la operación sea rápida y sin pérdida de calidad. TODO: Turn this 'copy' to AudioCodec and VideoCodec (?)

        Notation: **-c {codec}**
        """
        # We cannot control the big amount of video codecs that
        # are available so we allow any string as parameter

        # TODO: Validate provided 'codec'
        # TODO: This method has a variation, it can be '-c:a' or '-c:v'
        if not PythonValidator.is_instance(codec, [FfmpegVideoCodec, FfmpegAudioCodec]):
            try:
                codec = FfmpegVideoCodec.to_enum(codec)
            except:
                pass

        if not PythonValidator.is_instance(codec, [FfmpegVideoCodec, FfmpegAudioCodec]):
            try:
                codec = FfmpegAudioCodec.to_enum(codec)
            except:
                pass

        if PythonValidator.is_instance(codec, [FfmpegVideoCodec, FfmpegAudioCodec]):
            codec = codec.value

        return f'-c {codec}'
    
    @staticmethod
    def map(map: str):
        """
        Set input stream mapping.
        -map [-]input_file_id[:stream_specifier][,sync_file_id[:stream_s set input stream mapping

        # TODO: Improve this

        Notation: **-map {map}**
        """
        return f'-map {map}'
    
    @staticmethod
    def filter(filter: FfmpegFilter):
        """
        Sets the expected filter to be used.

        Notation: **-filter {filter}**
        """
        filter = FfmpegFilter.to_enum(filter).value

        return f'-filter {filter}'
    
    @staticmethod
    def frame_rate(frame_rate: int):
        """
        Sets the frame rate (Hz value, fraction or abbreviation)

        Notation: **-r {frame_rate}**
        """
        # TODO: Validate 'frame_rate'

        return f'-r {str(frame_rate)}'
    
    @staticmethod
    def pixel_format(format: FfmpegPixelFormat):
        """
        Set the pixel format.

        Notation: **-pix_fmt {format}**
        """
        format = FfmpegPixelFormat.to_enum(format).value

        return f'-pix_fmt {format}'
    
    @staticmethod
    def scale_with_size(size: tuple):
        """
        Set a new size.

        Notation: **-vf scale=size[0]:size[1]**
        """
        return f'-vf scale={str(int(size[0]))}:{str(int(size[1]))}'

    @staticmethod
    def scale_with_factor(w_factor: float, h_factor: float):
        """
        Set a new size multiplying by a factor.

        Notation: **-vf "scale=iw*w_factor:ih*h_factor"**
        """
        return f'-vf "scale=iw*{str(w_factor)}:ih*{str(h_factor)}"'

    @staticmethod
    def crop(size: tuple, origin: tuple):
        """
        Crop the video to a new with the provided 'size'
        starting with the top left corner at the given
        'origin' position of the original video.
        
        Notation: **-vf "crop=size[0]:size[1]:origin[0]:origin[1]"**
        """
        return f"-vf \"crop={str(int(size[0]))}:{str(int(size[1]))}:{str(int(origin[0]))}:{str(int(origin[1]))}\""
    
    @staticmethod
    def seeking(seconds: int):
        """
        Skip the necessary amount of time to go directly
        to the provided 'seconds' time of the input (that
        must be provided after this).

        Notation: **-ss 00:00:03**
        """
        return f'-ss {seconds_to_hh_mm_ss(seconds)}'
    
    @staticmethod
    def to(seconds: int):
        """
        Used with 'seeking' to match the duration we want
        to apply to the new trimmed input. In general, 
        this will be the amount of 'seconds' to be played.

        Notation: **-to 00:00:05**
        """
        return f'-to {seconds_to_hh_mm_ss(seconds)}'
    
class FfmpegCommand:
    """
    Class to represent a command to be built and
    executed by the FfmpegHandler.

    A valid example of a command is built like this:
    
    FfmpegCommand([
        FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
        FfmpegFlag.safe_routes(0),
        FfmpegFlag.overwrite,
        FfmpegFlag.frame_rate(frame_rate),
        FfmpegFlag.input(concat_filename),
        FfmpegFlag.video_codec(FfmpegVideoCodec.QTRLE),
        FfmpegFlag.pixel_format(pixel_format), # I used 'argb' in the past
        output_filename
    ])
    """
    args: list[Union[FfmpegFlag, any]] = None
    
    def __init__(self, args: list[Union[FfmpegFlag, any]]):
        # TODO: Validate args
        self.args = args

    def run(self):
        """
        Run the command.
        """
        run(self.__str__())

    def __str__(self):
        """
        Turn the command to a string that can be directly
        executed as a ffmpeg command.
        """
        # TODO: Clean args (?)
        # Remove 'None' args, our logic allows them to make it easier
        args = [arg for arg in self.args if arg is not None]

        return f"ffmpeg {' '.join(args)}"

class FfmpegHandler:
    """
    Class to simplify and encapsulate ffmpeg functionality.
    """
    @staticmethod
    def validate_video_filename(video_filename: str):
        # TODO: Validate and raise Exception if invalid
        pass

    @staticmethod
    def write_concat_file(filenames: str):
        """
        Writes the files to concat in a temporary text file with
        the required format and returns that file filename. This
        is required to use different files as input.
        """
        text = ''
        for filename in filenames:
            text += f"file '{filename}'\n"

        # TODO: Maybe this below is interesting for the 'yta_general_utils.file.writer'
        # open('concat.txt', 'w').writelines([('file %s\n' % input_path) for input_path in input_paths])
        filename = create_temp_filename('concat_ffmpeg.txt')
        FileWriter.write_file(text, filename)

        return filename

    @staticmethod
    def run_command(command: Union[list[FfmpegFlag, any], FfmpegCommand]):
        """
        Runs the provided ffmpeg 'command'.
        """
        if not PythonValidator.is_instance(command, FfmpegCommand):
            command = FfmpegCommand(command)

        command.run()

    # TODO: Check this one below
    @staticmethod
    def get_audio_from_video_deprecated(video_filename: str, codec: FfmpegAudioCodec = None, output_filename: str = None):
        """
        Exports the audio of the provided video to a local file named
        'output_filename' if provided or as a temporary file.

        # TODO: This has not been tested yet.

        This methods returns a tuple with the audio as a moviepy audio 
        in the first place and the filename of the file generated in
        the second place.
        """
        FfmpegHandler.validate_video_filename(video_filename)
        
        if not output_filename:
            output_filename = create_temp_filename('temp_audio.mp3')
        # TODO: Validate 'output_filename' is a valid audio filename

        if codec:
            codec = FfmpegAudioCodec.to_enum(codec)

        FfmpegCommand([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.audio_codec(codec) if codec else None,
            output_filename
        ]).run()

        # TODO: This was .to_audiofileclip() before,
        # remove this comment if working
        return AudioParser.as_audioclip(output_filename), output_filename
    
    @staticmethod
    def get_audio_from_video(video_filename: str, output_filename: str = None):
        """
        Exports the audio of the provided video to a local file named
        'output_filename' if provided or as a temporary file.

        This methods returns a tuple with the audio as a moviepy audio 
        in the first place and the filename of the file generated in
        the second place.
        """
        FfmpegHandler.validate_video_filename(video_filename)
        
        if not output_filename:
            output_filename = create_temp_filename('temp_audio.mp3')

        # TODO: Verify valid output_filename extension

        FfmpegCommand([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.map('0:1'),
            output_filename
        ]).run()

        # TODO: This was .to_audiofileclip() before,
        # remove this comment if working
        return AudioParser.as_audioclip(output_filename), output_filename

    @staticmethod
    def get_best_thumbnail(video_filename: str, output_filename: str = None):
        """
        Gets the best thumbnail of the provided 'video_filename'.

        This methods returns a tuple with the thumbnail as a pillow 
        image in the first place and the filename of the file generated
        in the second place.
        """
        FfmpegHandler.validate_video_filename(video_filename)

        if not output_filename:
            output_filename = create_temp_filename('temp_thumbnail.png')
        # TODO: Verify valid extension

        FfmpegCommand([
            FfmpegFlag.input(video_filename),
            FfmpegFlag.filter(FfmpegFilter.THUMBNAIL),
            output_filename
        ]).run()

        return ImageParser.to_pillow(output_filename), output_filename
    
    @staticmethod
    def concatenate_videos(video_filenames: str, output_filename: str = None):
        """
        Concatenates the provided 'video_filenames' in the order in
        which they are provided.

        This methods returns a tuple with the new video as a moviepy
        video in the first place and the filename of the file generated
        in the second place.
        """
        for video_filename in video_filenames:
            FfmpegHandler.validate_video_filename(video_filename)

        # TODO: Use new output handler
        if not output_filename:
            output_filename = create_temp_filename('concatenated_video.mp4')

        concat_filename = FfmpegHandler.write_concat_file(video_filenames)

        FfmpegCommand([
            FfmpegFlag.overwrite,
            FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
            FfmpegFlag.safe_routes(0),
            FfmpegFlag.input(concat_filename),
            FfmpegFlag.codec('copy'),
            output_filename
        ]).run()

        return VideoParser.to_moviepy(output_filename), output_filename
    
    @staticmethod
    def concatenate_images(image_filenames: str, frame_rate = 60, pixel_format: FfmpegPixelFormat = FfmpegPixelFormat.YUV420p, output_filename: str = None):
        """
        Concatenates the provided 'image_filenames' in the order in
        which they are provided.

        This methods returns a tuple with the new video as a moviepy
        video in the first place and the filename of the file generated
        in the second place.
        """
        for image_filename in image_filenames:
            FfmpegHandler.validate_video_filename(image_filename)

        if not output_filename:
            output_filename = create_temp_filename('concatenated_video.mp4')

        concat_filename = FfmpegHandler.write_concat_file(image_filenames)

        # TODO: Should we check the pixel format or give freedom (?)
        # pixel_format = FfmpegPixelFormat.to_enum(pixel_format)

        FfmpegCommand([
            FfmpegFlag.force_format(FfmpegVideoFormat.CONCAT),
            FfmpegFlag.safe_routes(0),
            FfmpegFlag.overwrite,
            FfmpegFlag.frame_rate(frame_rate),
            FfmpegFlag.input(concat_filename),
            FfmpegFlag.video_codec(FfmpegVideoCodec.QTRLE),
            FfmpegFlag.pixel_format(pixel_format), # I used 'argb' in the past
            output_filename
        ]).run()

        return VideoParser.to_moviepy(output_filename), output_filename

    @staticmethod
    def resize_video(video_filename: str, size: tuple, output_filename: Union[str, None] = None):
        """
        Resize the provided 'video_filename', by keeping
        the aspect ratio (cropping if necessary), to the
        given 'size' and stores it locally as
        'output_filename'.

        See more: 
        https://www.gumlet.com/learn/ffmpeg-resize-video/
        """
        if not PythonValidator.is_string(video_filename):
            raise Exception('The provided "video_filename" parameter is not a valid string.')
        
        if not PythonValidator.is_tuple(size):
            raise Exception('The provided "size" parameter is not a tuple.')
        
        if not PythonValidator.is_string(output_filename):
            raise Exception('The provided "output_filename" parameter is not a valid string.')
        
        if not FileValidator.file_is_video_file(video_filename):
            raise Exception('The provided "video_filename" is not a valid video file name.')
        
        if output_filename is None:
            output_filename = create_temp_filename('tmp_resized_ffmpeg.mp4')

        validate_size(size, 'size')

        w, h = get_video_size(video_filename)

        if (w, h) == size:
            # No need to resize, we just copy it to output
            FileHandler.copy_file(video_filename, output_filename)
        else:
            # First, we need to know if we need to scale it
            original_ratio = w / h
            new_ratio = size[0] / size[1]

            if original_ratio > new_ratio:
                # Original video is wider than the expected one
                new_size = w * (size[1] / h), size[1]
            elif original_ratio < new_ratio:
                # Original video is higher than the expected one
                new_size = size[0], h * (size[0] / w)
            else:
                new_size = size[0], size[1]

            tmp_filename = create_temp_filename('tmp_ffmpeg_scaling.mp4')

            # Scale to new dimensions
            FfmpegCommand([
                FfmpegFlag.input(video_filename),
                FfmpegFlag.scale_with_size(new_size),
                tmp_filename
            ]).run()

            # Now, with the new video resized, we look for the
            # cropping points we need to apply and we crop it
            top_left, _ = get_cropping_points_to_keep_aspect_ratio(new_size, size)

            # Second, we need to know if we need to crop it
            FfmpegCommand([
                FfmpegFlag.input(tmp_filename),
                FfmpegFlag.crop(size, top_left),
                FfmpegFlag.overwrite,
                output_filename
            ]).run()

        return output_filename
    
    @staticmethod
    def trim(video_filename: str, start_seconds: int, duration_seconds: int, output_filename: Union[str, None] = None):
        """
        Trims the provided 'video_filename' and generates a
        shorter resource that is stored as 'output_filename'.
        This new resource will start from 'start_seconds' and
        last the provided 'duration_seconds'.

        Thank you:
        https://www.plainlyvideos.com/blog/ffmpeg-trim-videos
        https://trac.ffmpeg.org/wiki/Seeking
        """
        if not PythonValidator.is_string(video_filename):
            raise Exception('The provided "video_filename" parameter is not a valid string.')
        
        if not NumberValidator.is_positive_number(start_seconds, do_include_zero = True):
            raise Exception('The provided "start_seconds" is a negative number.')

        if not NumberValidator.is_positive_number(duration_seconds, do_include_zero = True):
            raise Exception('The provided "duration_seconds" is a negative number.')
        
        if not output_filename:
            output_filename = create_temp_filename('tmp_trimmed_ffmpg.mp4')

        command = FfmpegCommand([
            FfmpegFlag.seeking(start_seconds),
            FfmpegFlag.input(video_filename),
            FfmpegFlag.to(duration_seconds),
            FfmpegFlag.codec(FfmpegVideoCodec.COPY),
            FfmpegFlag.overwrite,
            output_filename
        ])
        print(command)
        command.run()

        #ffmpeg_command = f'-ss 00:02:05 -i {video} -to 00:03:10 -c copy video-cutted-ffmpeg.mp4'
        return output_filename

    # TODO: This method must replace the one in 
    # yta_multimedia\video\audio.py > set_audio_in_video_ffmpeg
    @staticmethod
    def set_audio(video_filename: str, audio_filename: str, output_filename: Union[str, None] = None):
        if not PythonValidator.is_string(video_filename):
            raise Exception('The provided "video_filename" parameter is not a valid string.')
        
        if not PythonValidator.is_string(audio_filename):
            raise Exception('The provided "audio_filename" parameter is not a valid string.')
        
        if not PythonValidator.is_string(output_filename):
            raise Exception('The provided "output_filename" parameter is not a valid string.')
        
        if not FileValidator.file_is_audio_file(audio_filename):
            raise Exception('The provided "audio_filename" is not a valid audio file name.')
        
        if not FileValidator.file_is_video_file(video_filename):
            raise Exception('The provided "video_filename" is not a valid video file name.')

        if not output_filename:
            output_filename = create_temp_filename('tmp_audio_ffmpg.mp4')
        
        # cls.run_command([
        #     FfmpegFlag.input(video_filename),
        #     FfmpegFlag.input(audio_filename),
        #     output_filename
        # # TODO: Unfinished
        # ])

        # TODO: Is this actually working (?)
        run(f"ffmpeg -i {video_filename} -i {audio_filename} -c:v copy -c:a aac -strict experimental -y {output_filename}")
        
        # Apparently this is the equivalent command according
        # to ChatGPT, but maybe it doesn't work
        # ffmpeg -i input_video -i input_audio -c:v copy -c:a aac -strict experimental -y output_filename

        # There is also a post that says this:
        # ffmpeg -i input.mp4 -i input.mp3 -c copy -map 0:v:0 -map 1:a:0 output.mp4
        # in (https://superuser.com/a/590210)


        # # TODO: What about longer audio than video (?)
        # # TODO: This is what was being used before FFmpegHandler
        # input_video = ffmpeg.input(video_filename)
        # input_audio = ffmpeg.input(audio_filename)

        # ffmpeg.concat(input_video, input_audio, v = 1, a = 1).output(output_filename).run(overwrite_output = True)

    
    # TODO: Create a 'set_audio_in_video' method to replace the
    # yta_multimedia\video\audio.py > set_audio_in_video_ffmpeg

    # TODO: Keep going

    # https://www.reddit.com/r/ffmpeg/comments/ks8zfs/comment/gieu7x6/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
    # https://stackoverflow.com/questions/38368105/ffmpeg-custom-sequence-input-images/51618079#51618079
    # https://stackoverflow.com/a/66014158


"""
I'm placing the edition settings here until I find
a better place, but I want to keep them in the same
file by now to avoid duplicating them in the files
I am building for new concepts.

TODO: Move this to a definitive file in a near 
future, please
"""
from yta_image.edition.settings import COLOR_TEMPERATURE_CHANGE_LIMIT, BRIGHTNESS_LIMIT, CONTRAST_LIMIT, SHARPNESS_LIMIT, WHITE_BALANCE_LIMIT, SPEED_FACTOR_LIMIT


MAX_TIMELINE_LAYER_DURATION = 1200
"""
The maximum duration a timeline layer can have
according to all the subclips on it. This value
can change to allow longer timeline layers.
"""

LAYERS_INDEXES_LIMIT = (0, 9)
"""
The limit of the layers indexes we have, starting
from 0, so only the upper limit + 1 layers are
available in the edition system.
"""
VOLUME_LIMIT = (0, 300)
"""
The limit of the volumen adjustment. Zero (0) means
silence, 100 means the original audio volume and
300 means 3 times higher volume.
"""
ZOOM_LIMIT = (1, 500)
"""
The limit of the zoom adjustment. One (1) means a
zoom out until the video is a 1% of its original 
size, 100 means the original size and 500 means a
zoom in to reach 5 times the original video size.
"""
ROTATION_LIMIT = (-360, 360)
"""
The limit of the rotation adjustment. Zero (0) 
means no rotation, while 90 means rotated 90 
degrees to the left, 180 means flipped vertically
and 360 means no rotation.
"""
COLOR_TEMPERATURE_LIMIT = COLOR_TEMPERATURE_CHANGE_LIMIT
"""
The limit of the color temperature adjustment. Zero
(0) means no change, while -50 means...

TODO: Fulfill this as it depends on another const
"""
BRIGHTNESS_LIMIT = BRIGHTNESS_LIMIT
"""
The limit of the image brightness adjustment.
"""
CONTRAST_LIMIT = CONTRAST_LIMIT
"""
The limit of the image contrast adjustment.
"""
SHARPNESS_LIMIT = SHARPNESS_LIMIT
"""
The limit of the image sharpness adjustment.
"""
WHITE_BALANCE_LIMIT = WHITE_BALANCE_LIMIT
"""
The limit of the image white balance adjustment.
"""
SPEED_FACTOR_LIMIT = SPEED_FACTOR_LIMIT
"""
The limit of the speed factor adjustment.
"""


from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from abc import abstractmethod

import numpy as np


class PositionMovement:
    """
    Handles a movement from the 'initial_position' to
    the 'final_position'
    """
    def __init__(self, initial_position: tuple[int, int], final_position: tuple[int, int]):
        # TODO: Validate 'initial_position' is valid
        # TODO: Validate 'final_position' is valid

        self.initial_position = initial_position
        self.final_position = final_position

    @abstractmethod
    def get_position(self, n: float, rate_function: RateFunctionArgument = RateFunctionArgument.default()):
        """
        Obtain the position on the given 'n' moment of
        the movement ('n' value must be a normalized
        value between 0 and 1 representing the amount
        of movement done for the moment to be
        calculated where 0 is the start and 1 the end).
        """
        pass

class LinearPositionMovement(PositionMovement):
    """
    Handles a linear movement from the 'initial_position'
    to the 'final_position'.
    """
    def get_position(self, n: float, rate_function: RateFunctionArgument = RateFunctionArgument.default()):
        return linear(self.initial_position, self.final_position, n, rate_function)
    
class SinusoidalPositionMovement(PositionMovement):
    """
    Handles a sinusoidal movement from the 'initial_position'
    to the 'final_position'.
    """
    def get_position(self, n: float, rate_function: RateFunctionArgument = RateFunctionArgument.default()):
        return sinusoidal(self.initial_position, self.final_position, n, rate_function)
    
class ZigZagPositionMovement(PositionMovement):
    """
    Handles a zigzag movement from the 'initial_position'
    to the 'final_position'.
    """
    def get_position(self, n: float, rate_function: RateFunctionArgument = RateFunctionArgument.default()):
        return zigzag(self.initial_position, self.final_position, n, rate_function) 

def linear(
    initial_position: tuple[int, int],
    final_position: tuple[int, int],
    n: float,
    rate_function: RateFunctionArgument = RateFunctionArgument.default()
):
    # TODO: Validate 'initial_position' is valid
    # TODO: Validate 'final_position' is valid
    
    # Obtain the real factor according to the rate function
    n = rate_function.get_n_value(n)

    return (
        initial_position[0] + (final_position[0] - initial_position[0]) * n,
        initial_position[1] + (final_position[1] - initial_position[1]) * n,
    )

def sinusoidal(
    initial_position: tuple,
    final_position: tuple,
    n: float,
    rate_function: RateFunctionArgument = RateFunctionArgument.default()
):
    # TODO: Validate 'initial_position' is valid
    # TODO: Validate 'final_position' is valid

    WAVE_AMPLITUDE = 100
    WAVE_FREQUENCY = 2

    # Obtain the real factor according to the rate function
    n = rate_function.get_n_value(n)

    return (
        initial_position[0] + n * (final_position[0] - initial_position[0]),
        initial_position[1] + n * (final_position[1] - initial_position[1]) + WAVE_AMPLITUDE * np.sin(2 * np.pi * WAVE_FREQUENCY * n)
    )

def zigzag(
    initial_position: tuple[int, int],
    final_position: tuple[int, int],
    n: float,
    rate_function: RateFunctionArgument = RateFunctionArgument.default()
):
    # TODO: Validate 'initial_position' is valid
    # TODO: Validate 'final_position' is valid

    AMPLITUDE = 15

    # Obtain the real factor according to the rate function
    n = rate_function.get_n_value(n)

    return (
        initial_position[0] + n * (final_position[0] - initial_position[0]),
        initial_position[1] + n * (final_position[1] - initial_position[1]) + AMPLITUDE * np.sin(10 * np.pi * n)
    )


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from moviepy import Clip
from moviepy.video.fx import BlackAndWhite


class BlackAndWhiteEffect(Effect):
    """
    This effect will make the clip appear in black and
    white colors.
    """
    def apply(self, video: Clip) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        return BlackAndWhite().apply(video)
    
    # TODO: I don't need this
    def apply_over_video(self, video, background_video):
        return super().apply_over_video(video, background_video)